//
// cbAudioCOntroller.m
//  CB
//
//  Created by Stanko Juzbasic on 10/9/15.
//  Copyright (c) 2015 Stanko Juzbasic. All rights reserved.
//

#import "cbAudioController.h"

#define kInputBus 1
#define kOutputBus 0
#define kBufferLength 2048
#define kBufferOrder  12
#define twopi    6.283185307179586476925287
#define TWOPI    6.283185307179586476925287

 
@implementation cbAudioController
//_____________________________________________________________________________
// generic result checker
//_____________________________________________________________________________
static int checkResultLite(OSStatus error, const char *operation)
{   
	if (error == noErr) return 0;
	
	char str[20];

	// see if it appears to be a 4-char-code
	*(UInt32 *)(str + 1) = CFSwapInt32HostToBig(error);
	if (isprint(str[1]) && isprint(str[2]) && isprint(str[3]) && isprint(str[4])) {
		str[0] = str[5] = '\'';
		str[6] = '\0';
	} else
		// no, format as an integer
		sprintf(str, "%d", (int)error);
    return 1;
}
//_____________________________________________________________________________
// audio setup & initialization
//_____________________________________________________________________________
- (void)createToneUnit
{
    UInt32 size = sizeof(Float64);
    UInt32 enableOutput = 1;
    
    
	// Configure the search parameters to find the default I/O unit
	// kAudioUnitSubType_DefaultOutput on Mac OS X for output only
	// kAudioUnitSubType_HALOutput on Mac OS X to also capture input signal!
	AudioComponentDescription audioCompDesc;
	audioCompDesc.componentType = kAudioUnitType_Output;
	audioCompDesc.componentSubType = kAudioUnitSubType_HALOutput;
	audioCompDesc.componentManufacturer = kAudioUnitManufacturer_Apple;
	audioCompDesc.componentFlags = 0;
	audioCompDesc.componentFlagsMask = 0;
    
	
	// Check for the described input-enableable output unit
	AudioComponent halOutput = AudioComponentFindNext(NULL, &audioCompDesc);
	NSAssert(audioRenderCallback, @"Couldn't find unit of subType HAL output");
	
	// Create a new unit based on this that we'll use for I/O
    checkResultLite(AudioComponentInstanceNew(halOutput, &toneUnit),
			   "Couldn't get audio unit instance");
    
    //enable output
    checkResultLite(AudioUnitSetProperty (toneUnit,
                                     kAudioOutputUnitProperty_EnableIO,
                                     kAudioUnitScope_Output,
                                     kOutputBus,
                                     &enableOutput,
                                     sizeof(enableOutput)),
			   "Couldn't enable output");
    
    
    //get sample rate
    checkResultLite(AudioUnitGetProperty(toneUnit,
                                    kAudioUnitProperty_SampleRate,
                                    kAudioUnitScope_Output,
                                    kOutputBus,
                                    &R,
                                    &size),
               "Couldn't get sample rate");

	// Set tone rendering function on the unit 
	AURenderCallbackStruct input;
	input.inputProc = audioRenderCallback;
	input.inputProcRefCon = self;
    
   
    checkResultLite( AudioUnitSetProperty(toneUnit,
                                     kAudioUnitProperty_SetRenderCallback,
                                     kAudioUnitScope_Input,
                                     kOutputBus,
                                     &input,
                                     sizeof(input)),
               "Error setting input callback: ");
    
    UInt32 flag = 0;
    checkResultLite( AudioUnitSetProperty(toneUnit,
                                  kAudioUnitProperty_ShouldAllocateBuffer,
                                  kAudioUnitScope_Output,
                                  kInputBus,
                                  &flag,
                                  sizeof(flag)),
                "Error unsetting buffer:");    
	
	// Set streamFormat to 32 bit, single channel, floating point, linear PCM
	const int four_bytes_per_float = 4;
	const int eight_bits_per_byte = 8;
	streamFormat.mSampleRate = R;
	streamFormat.mFormatID = kAudioFormatLinearPCM;
	streamFormat.mFormatFlags =
    kAudioFormatFlagsNativeFloatPacked | kAudioFormatFlagIsNonInterleaved;
	streamFormat.mBytesPerPacket = four_bytes_per_float;
	streamFormat.mFramesPerPacket = 1;
	streamFormat.mBytesPerFrame = four_bytes_per_float;
	streamFormat.mChannelsPerFrame = 1;
	streamFormat.mBitsPerChannel = four_bytes_per_float * eight_bits_per_byte;
    
    //set streamFormat for input scope
	checkResultLite(AudioUnitSetProperty (toneUnit,
                                     kAudioUnitProperty_StreamFormat,
                                     kAudioUnitScope_Input,
                                     kOutputBus,
                                     &streamFormat,
                                     sizeof(AudioStreamBasicDescription)),
               "Couldn't set StreamFormat on input scope / kOutputBus");

    //set global variables 
    self->sampleRate = R;
    self->f = 523. ;
    self->phase = 0.;
    self->previous_phase_increment = 0.;
    self->state = idle;
    self->T     = (int)self->_segmControl.selectedSegment;
    if(debug)NSLog(@"T = %d\n", self->T);
    
    checkResultLite(AudioUnitInitialize(toneUnit),
               "Error initializing unit:");
    checkResultLite(AudioOutputUnitStart(toneUnit),
               "Error starting unit: ");
}
//_____________________________________________________________________________
//  initialize/uninitialize everything
//_____________________________________________________________________________
- (void)setup {
    debug = false;
    // Create the semaphore
    mSemaphore = dispatch_semaphore_create(0);
    // Setup audio
    [self createToneUnit];
    
    // Initialize buffer
    TPCircularBufferInit(&(self->buffer), kBufferLength);
    // Initialize cbuffer
    self->cbuffer = *cbuf_new(12);
    //sbuffer
    if(SatRingCreate (&(self->sbuffer),  512, 1, 4) != 0)
        NSLog(@"\n\Error: couldn't create sbuffer\n");
    if(SatRingInit (&(self->sbuffer)) != 0)
        NSLog(@"\n\Error: couldn't initizlize sbuffer\n");
/**/

    NSLog(@"\n\tAudio SET UP\n");    
}
//_____________________________________________________________________________
- (void)dealloc {
    
    // Release buffer resources
    //SatRingDestroy (&(self->sbuffer));
    //cbuf_free(&(self->cbuffer)); //this crashes (!)
    //TPCircularBufferCleanup(&(self->buffer));
    // Clean up semaphore
    dispatch_release(mSemaphore);
    if(toneUnit){
        AudioOutputUnitStop(toneUnit);
		AudioUnitUninitialize(toneUnit);
		AudioComponentInstanceDispose(toneUnit);
        NSLog(@"\n\tAudio RELEASED\n");
    }
    [super dealloc];
}
//_____________________________________________________________________________
//  actions from GUI
//_____________________________________________________________________________
- (IBAction)sliderDragged:(NSSlider *)sender
{
    self->f = updateFrequency(sender.floatValue);
}
//_____________________________________________________________________________
Float32 updateFrequency(float sliderValue)
{
    return (Float32)261.5*(powf(2.,sliderValue));
}
//_____________________________________________________________________________
- (IBAction)playButtonPressed:(NSButton *)sender {
    Boolean on = false;
    if(sender.state == NSOnState){
        if(debug)NSLog(@"\nON\n");
        on = true;
        if (self->state == idle){
        self->state = starting;
        [self dproducer: on];
            //self.bufferButton.enabled = false;
            self.segmControl.enabled = false;
        }
            
    }else{
        if (self->state == active){
        self->state = stopping;
        if(debug)NSLog(@"OFF\n\n");
            //self.bufferButton.enabled = true;
            self.segmControl.enabled  = true;
        }
    }
    return;
}
//____________________________________________________________________________
- (IBAction)transparentButtonPressed:(NSButton *)sender
{
    debug = !debug;
}
//_____________________________________________________________________________
- (IBAction)segmControlPressed:(NSSegmentedControl *)sender
{
    NSInteger pressedSegment = [sender selectedSegment];
    switch (pressedSegment) {
        case 0:
            self->T = 0;
            break;
        case 1:
            self->T = 1;
            break;
        case 2:
            self->T = 2;
            break;
        default:
            break;
    }
    
}
//_____________________________________________________________________________
//  dispatcher (producer queue)
//_____________________________________________________________________________
- (void) dproducer: (Boolean) on
{
    if (on == true)
    {
        dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_HIGH, 0), ^{
            [self producer: on];
        });
    }    
    return;
}
//_____________________________________________________________________________
//  updateable producer routine
//  it writes to any of the three available circular buffers
//_____________________________________________________________________________
- (void)producer: (Boolean)on
{    
    int     bytesToCopy;    
    Float32 samples[self->frames];
    Float32 phi = self->phase;
    Float32 wait_interval;

    while(on){
        availableBytes = 0;
        bytesToCopy = 0;
        wait_interval = 1.5*(self->frames)/(self->sampleRate);
        Float32 cycle= self->sampleRate/self->f;
        Float32 phase_increment = twopi/cycle;
        Float32 dphi = (phase_increment - self->previous_phase_increment)/(Float32)(self->frames);
        Float32 damp = 1./(Float32)(self->frames);
        for (int i = 0; i < self->frames; i++) {
            samples[i] = .5*sinf(phi);
            bytesToCopy += sizeof(Float32);
            phi = phi + self->previous_phase_increment + i*dphi;
            if (phi > twopi) phi -= twopi;
        }
        if(self->state == idle)    memset(samples, 0, self->frames*sizeof(Float32));
        if(self->state == stopped) memset(samples, 0, self->frames*sizeof(Float32));
        if(self->state == starting){
            for(int i = 0; i < self->frames; i++)samples[i] *= i*damp;
            self->state = active;
        }
        if(self->state == stopping){
            for(int i = 0; i < self->frames; i++)samples[i] *= (self->frames - i)*damp;
            self->state = stopped; 
        }
                
        self->phase = phi;
        self->previous_phase_increment = phase_increment;
        /**/
        //TPC
        if(self->T == 1){
        float *head = TPCircularBufferHead(&(self->buffer), &availableBytes);        
        memcpy(head,(const void*)samples,MIN(bytesToCopy, availableBytes));//copy samples to head
        TPCircularBufferProduce(&(self->buffer),MIN(bytesToCopy,availableBytes));
        //end TPC
        }
        
        // CB
        //int unused = cbuf_unusedspace(&(self->cbuffer));
        if(self->T == 2){
        int numBytes = cbuf_offer(&(self->cbuffer), (unsigned char*)samples, bytesToCopy);
        if(numBytes > 0 && debug) NSLog(@"numBytes = %d\n", numBytes);
        }
        // SRB
        SatRingWriteFrames (&(self->sbuffer), (uint8_t*)samples, self->frames );
        //end SRB
       
        // Wait for a signal for wait_interval (deduced experimentally)
        dispatch_semaphore_wait(mSemaphore, dispatch_time(DISPATCH_TIME_NOW, wait_interval * NSEC_PER_SEC));
        if(self->state == stopped){self->state = idle; on = false;}
            
    }
    if(debug)NSLog(@"RETURNING\n");
    return;
}
//_____________________________________________________________________________
//  render callback (consumer thread)
//  it reads from any of the three available circular buffers
//_____________________________________________________________________________
static OSStatus audioRenderCallback(void *inRefCon,
                                      AudioUnitRenderActionFlags *ioActionFlags,
                                      const AudioTimeStamp *inTimeStamp,
                                      UInt32 inBusNumber,
                                      UInt32 inNumberFrames,
                                      AudioBufferList *ioData) {    

    cbAudioController *THIS = (cbAudioController *)inRefCon;
    // An event happens in the render thread- signal whoever is waiting
    if ((THIS->state == active)||(THIS->state == stopping)||(THIS->state == starting))
        dispatch_semaphore_signal(mSemaphore);
    
    // Mono tone generator: we only need one buffer
    const int channel = 0;
    Float32* targetBuffer = (Float32 *)ioData->mBuffers[channel].mData;
    memset(targetBuffer,0,inNumberFrames*sizeof(Float32));

    // Pull samples from circular buffer(s)

    int32_t availableBytes, sampleCount = 0;
    int bytesToCopy = ioData->mBuffers[0].mDataByteSize;
    
    // TPC
    if(THIS->T == 1){
    Float32 *buffer = TPCircularBufferTail(&THIS->buffer, &availableBytes);
    memcpy(targetBuffer, (const void*)buffer, MIN(bytesToCopy, availableBytes));
    sampleCount = availableBytes/sizeof(Float32);
    TPCircularBufferConsume(&THIS->buffer, availableBytes);
    //end TPC
    // SRB
    }else if(THIS->T == 0){
        if(THIS->state == idle){
            memset(targetBuffer,0,inNumberFrames*sizeof(Float32));
        }else SatRingRead (&THIS->sbuffer, (uint8_t*)targetBuffer );
    //end SRB
    // CB
    }else if(THIS->T == 2){
     
        if(THIS->state == idle)
            memset(targetBuffer,0,inNumberFrames*sizeof(Float32));
        else{
            int used = cbuf_usedspace(&(THIS->cbuffer));
            unsigned char *ptr0 = cbuf_peek(&(THIS->cbuffer));
            if(ptr0 != NULL)memcpy(targetBuffer, (const void*)ptr0, used);
            cbuf_poll(&(THIS->cbuffer), used);
            if(debug)NSLog(@"usedBytes= %d\n",used);
            
        }
    // end CB
    }
    
    THIS->frames = inNumberFrames;

    if(debug){
    NSLog(@"\n\tsampleCount = %d\n",sampleCount);
    NSLog(@"\n\tinNumberFrames = %d\n",inNumberFrames);    
    }
    return noErr;
}
//_____________________________________________________________________________

@end
